#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mem.h>
#include <dir.h>

#include <allegro.h>

#define WORD  short int
#define UBYTE unsigned char

#ifndef FALSE
   #define FALSE 0
#endif
#ifndef TRUE
   #define TRUE -1
#endif

// reducing size of executable, for the allegro 3.9.34 (wip) lib

BEGIN_COLOR_DEPTH_LIST
   COLOR_DEPTH_8
END_COLOR_DEPTH_LIST

BEGIN_MIDI_DRIVER_LIST
   MIDI_DRIVER_SB_OUT
END_MIDI_DRIVER_LIST

BEGIN_DIGI_DRIVER_LIST
   DIGI_DRIVER_SB
END_DIGI_DRIVER_LIST

typedef struct
{
   WORD  xpos;
   WORD  ypos;
   WORD  unknown1;
   int   xgrid;
   int   ygrid;
   WORD  format;
   long  data_length;
   WORD  unknown2;
   long  data_offset;
   UBYTE * data;
} SUB_TILE_S;

typedef struct
{
   long       direction;
   long       flags;
   long       ysize;
   long       xsize;
   long       zeros1;
   long       orientation;
   long       main_index;
   long       sub_index;
   long       frame;
   long       unknown2;
   int        sub_tiles_flags[32];
   long       tiles_ptr;
   long       tiles_length;
   long       tiles_number;
   int        zeros2[12];
   SUB_TILE_S * tile;
} BLOCK_S;

// global datas
PALETTE the_pal;
BLOCK_S * glb_block[32];
long    glb_block_number[32];
int     glb_dt1_file = 0;
struct
{
   unsigned char p1;
   unsigned char p2;
   unsigned char p3;
   unsigned char p4;
} layer[12][128][128];


// ==========================================================================
// atexit() function, when the program end, to free the allocated memory
void ds1layer_exit(void)
{
   BLOCK_S     * block_ptr;
   SUB_TILE_S  * tile_ptr;
   int         b, t, i;

   for (i=0; i<32; i++)
   {
      if (glb_block[i] == NULL)
         return;
      for (b=0; b<glb_block_number[i]; b++)
      {
         block_ptr = glb_block[i] + b;
         if (block_ptr->tile != NULL)
         {
            for (t=0; t<block_ptr->tiles_number; t++)
            {
               tile_ptr = block_ptr->tile + t;
               if (tile_ptr->data != NULL)
                  free(tile_ptr->data);
            }
            free(block_ptr->tile);
         }
      }
      free(glb_block[i]);
   }
}


// ==========================================================================
// load a diablo 2 palette
void load_palette(char * d2pal)
{
   FILE * in;
   int  i;
   
   in = fopen(d2pal, "rb");
   if (in == NULL)
   {
      set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);
      printf("\ncan't open palette : %s\n", d2pal);
      exit(1);
   }
   for (i=0; i<256; i++)
   {
      // order is BGR, not RGB
      the_pal[i].b = fgetc(in) >> 2;
      the_pal[i].g = fgetc(in) >> 2;
      the_pal[i].r = fgetc(in) >> 2;
      /*
         >> 2 is needed because the internal range of the RGB componants
         in allegro (8 bpp mode) is from 0 to 63 (because of the mode 13h),
         not 0 to 255. Since this program only output bitmaps for the purpose
         of tests, and don't create dt1 files, it's ok.
      */
   }

   // make a grey background
   the_pal[0].r = 24;
   the_pal[0].g = 24;
   the_pal[0].b = 24;
   fclose(in);
}


// ==========================================================================
// read the dt1 main header
void read_dt1_header(FILE * in, int block_idx)
{
   long x1, x2, fsize;
   int  i, c, is_dt1 = TRUE;

   // file size
   fseek(in, 0, SEEK_END);
   fsize = ftell(in);
   fseek(in, 0, SEEK_SET);
   
   // signature
   fread(&x1, 1, 4, in);
   fread(&x2, 1, 4, in);
   if ((x1 != 7) || (x2 != 6))
      is_dt1 = FALSE;
      
   // zeros
   for (i=0; i<260; i++)
   {
      c = fgetc(in);
      if (c != 0)
         is_dt1 = FALSE;
   }
   
   // # of blocks
   fread(&glb_block_number[block_idx], 1, 4, in);
   if (glb_block_number[block_idx] < 0)
   {
      fclose(in);
      printf("error : number of blocks is negative (%li)\n", glb_block_number[block_idx]);
      exit(1);
   }

   // start of block header, always 0x114
   fread(&x1, 1, 4, in);
   if (x1 != 0x114)
      is_dt1 = FALSE;
   if (x1 >= fsize)
   {
      fclose(in);
      printf("error : start of blocks (%li) >= size of file (%li)\n", x1, fsize);
      exit(1);
   }

   // summary
   if (is_dt1 != TRUE)
{
      printf("possibly not a dt1 file, stop\n");
      exit(1);
}

   // get mem for block headers
   glb_block[block_idx] = (BLOCK_S *) malloc(sizeof(BLOCK_S) * glb_block_number[block_idx]);
   if (glb_block[block_idx] == NULL)
   {
      fclose(in);
      printf("not enough memory for %li blocks header\n", glb_block_number[block_idx]);
      exit(1);
   }
   memset(glb_block[block_idx], 0, sizeof(BLOCK_S) * glb_block_number[block_idx]);

   // ready to read block headers
   fseek(in, x1, SEEK_SET);
}


// ==========================================================================
// read the blocks headers
void read_blocks_headers(FILE * in, int block_idx)
{
   BLOCK_S * block_ptr;
   int     h, i;

   block_ptr = glb_block[block_idx];
   for (h=0; h<glb_block_number[block_idx]; h++)
   {
//fprintf(stderr, "block header %i\n", h);
      // direction
      fread(&block_ptr->direction, 1, 4, in);

      // flags
      fread(&block_ptr->flags, 1, 4, in);
      
      // ysize and xsize
      // negate the ysize : it is negative and I love when they are positive
      fread(&block_ptr->ysize, 1, 4, in);
      fread(&block_ptr->xsize, 1, 4, in);

      // zeros ?
      fread(&block_ptr->zeros1, 1, 4, in);
      
      // num[0] to num[3] (still unknown)
      fread(&block_ptr->orientation, 1, 4, in);
      fread(&block_ptr->main_index,  1, 4, in);
      fread(&block_ptr->sub_index,   1, 4, in);
      fread(&block_ptr->frame,       1, 4, in);

      // unknown 2 (always FF 00 FF 00 ?)
      fread(&block_ptr->unknown2, 1, 4, in);

      // sub-tiles flags (space for 32 max, but 25 in floor tiles)
      for (i=0; i<32; i++)
          block_ptr->sub_tiles_flags[i] = fgetc(in);

      // pointer to sub-tiles headers
      fread(&block_ptr->tiles_ptr,     1, 4, in);

      // sub-tiles length
      fread(&block_ptr->tiles_length,  1, 4, in);

      // # of sub-tiles (32 max ?)
      fread(&block_ptr->tiles_number,  1, 4, in);

      // zeros ?
      for (i=0; i<12; i++)
          block_ptr->zeros2[i] = fgetc(in);

      // ready to read next block header
      block_ptr++;
   }
}


// ==========================================================================
// read the sub-tiles of 1 block
void read_tiles_of_block(FILE * in, long b, int block_idx)
{
   BLOCK_S    * block_ptr;
   SUB_TILE_S * tile_ptr;
   UBYTE      * ptr;
   long       nb_tiles, t, f_pos;
   int        i;

//fprintf(stderr, "block gfx %i\n", b);

   // get ready
   block_ptr = glb_block[block_idx] + b;
   nb_tiles = block_ptr->tiles_number;

   // reserve memory
   block_ptr->tile = (SUB_TILE_S *) malloc(sizeof(SUB_TILE_S) * nb_tiles);
   if (block_ptr->tile == NULL)
   {
      fclose(in);
      printf("in block %li, not enough memory for %li tiles\n", b, nb_tiles);
      exit(1);
   }
   memset(block_ptr->tile, 0, sizeof(SUB_TILE_S) * nb_tiles);

   // ready to read tiles header
   fseek(in, block_ptr->tiles_ptr, SEEK_SET);
   tile_ptr = block_ptr->tile;

   // read each of them
   for (t=0; t<nb_tiles; t++)
   {
      // xpos & ypos
      fread(&tile_ptr->xpos, 1, 2, in);
      fread(&tile_ptr->ypos, 1, 2, in);

      // unknwon 1
      fread(&tile_ptr->unknown1, 1, 2, in);

      // xgrid & ygrid
      tile_ptr->xgrid = fgetc(in);
      tile_ptr->ygrid = fgetc(in);

      // sub-tile format
      fread(&tile_ptr->format, 1, 2, in);

      // sub-tile data length
      fread(&tile_ptr->data_length, 1, 4, in);

      // unknown 2
      fread(&tile_ptr->unknown2, 1, 2, in);
      
      // sub-til data offset
      fread(&tile_ptr->data_offset, 1, 4, in);

      // ready to read next sub-tile
      tile_ptr++;
   }

   // after the sub-tiles headers, their datas
   
   // read data of each sub-tiles
   tile_ptr = block_ptr->tile;
   for (t=0; t<nb_tiles; t++)
   {
      // file pointer = start of sub-tiles header + sub-tile data offset
      f_pos = block_ptr->tiles_ptr + tile_ptr->data_offset;
      fseek(in, f_pos, SEEK_SET);

      // reserved memory for data
      tile_ptr->data = (UBYTE *) malloc(tile_ptr->data_length);
      if (tile_ptr->data == NULL)
      {
         fclose(in);
         printf("in block %li, sub-tile %li, not enough memory for %li bytes\n",
            b, t, tile_ptr->data_length);
         exit(1);
      }

      // read data
      ptr = tile_ptr->data;
      for (i=0; i<tile_ptr->data_length; i++)
      {
         * ptr = fgetc(in);
         ptr++;
      }

      // next sub-tile data
      tile_ptr++;
   }
}


// ==========================================================================
// draw a 3d-isometric sub-tile
void draw_sub_tile_isometric (BITMAP * dst, int x0, int y0, const UBYTE * data, int length)
{
   UBYTE * ptr = data;
   int   x, y=0, n,
         xjump[15] = {14, 12, 10, 8, 6, 4, 2, 0, 2, 4, 6, 8, 10, 12, 14},
         nbpix[15] = {4, 8, 12, 16, 20, 24, 28, 32, 28, 24, 20, 16, 12, 8, 4};

   // 3d-isometric subtile is 256 bytes, no more, no less
   if (length != 256)
      return;

   // draw
   while (length > 0)
   {
      x = xjump[y];
      n = nbpix[y];
      length -= n;
      while (n)
      {
         putpixel(dst, x0+x, y0+y, * ptr);
         ptr++;
         x++;
         n--;
      }
      y++;
   }
}


// ==========================================================================
// draw a normal sub-tile (can be transparent, so there are "jump" coded)
void draw_sub_tile_normal (BITMAP * dst, int x0, int y0, const UBYTE * data, int length)
{
   UBYTE * ptr = data, b1, b2;
   int   x=0, y=0;

   // draw
   while (length > 0)
   {
      b1 = * ptr;
      b2 = * (ptr + 1);
      ptr += 2;
      length -= 2;
      if (b1 || b2)
      {
         x += b1;
         length -= b2;
         while (b2)
         {
            putpixel(dst, x0+x, y0+y, * ptr);
            ptr++;
            x++;
            b2--;
         }
      }
      else
      {
         x = 0;
         y++;
      }
   }
}


// ==========================================================================
/*
void draw_block(BITMAP * dst, int x0, int y0, int b, int block_idx)
{
   BLOCK_S    * block_ptr;
   SUB_TILE_S * tile_ptr;
   UBYTE      * ptr;
   int        t, x, y, length;

   block_ptr = glb_block[block_idx] + b;
   tile_ptr  = block_ptr->tile;

   for(t=0; t<block_ptr->tiles_number; t++)
   {
      length = tile_ptr->data_length;
      ptr    = tile_ptr->data;
      x      = tile_ptr->xpos;
      y      = tile_ptr->ypos;
      if (tile_ptr->format == 0x0001)
         draw_sub_tile_isometric(dst, x0+x, y0+y, ptr, length);
      else
         draw_sub_tile_normal(dst, x0+x, y0+y, ptr, length);
      tile_ptr++;
   }
}
*/

// ==========================================================================
void draw_block_search1(BITMAP * dst, int x0, int y0, int sub_idx, int main_idx)
{
   BLOCK_S    * block_ptr;
   SUB_TILE_S * tile_ptr;
   UBYTE      * ptr;
   int        b, t, x, y, length, dt1;


for (dt1=glb_dt1_file - 1; dt1 >= 0; dt1--)
{
   for (b=0; b<glb_block_number[dt1]; b++)
   {
      block_ptr = glb_block[dt1] + b;
      if ( (block_ptr->sub_index   == sub_idx)  &&
           (block_ptr->orientation == 0)        &&
           (block_ptr->main_index  == main_idx)
         )
      {
         if (block_ptr->orientation != 0)
            y0 += 80;
         tile_ptr  = block_ptr->tile;
         for(t=0; t<block_ptr->tiles_number; t++)
         {
            length = tile_ptr->data_length;
            ptr    = tile_ptr->data;
            x      = tile_ptr->xpos;
            y      = tile_ptr->ypos;
            if (tile_ptr->format == 0x0001)
               draw_sub_tile_isometric(dst, x0+x, y0+y, ptr, length);
            else
               draw_sub_tile_normal(dst, x0+x, y0+y, ptr, length);
            tile_ptr++;
         }
fprintf(stderr, "%c", b < 223 ? 32 + b : '*');
         return;
      }
   }
}
fprintf(stderr, ".");
}


// ==========================================================================
void draw_block_search2(BITMAP * dst, int x0, int y0, int orient, int sub_idx, int main_idx)
{
   BLOCK_S    * block_ptr;
   SUB_TILE_S * tile_ptr;
   UBYTE      * ptr;
   int        b, t, x, y, length, dt1;


for (dt1=glb_dt1_file - 1; dt1 >= 0; dt1--)
{
   for (b=0; b<glb_block_number[dt1]; b++)
   {
      block_ptr = glb_block[dt1] + b;
      if ( (block_ptr->sub_index   == sub_idx)  &&
           (block_ptr->orientation == orient)   &&
           (block_ptr->main_index  == main_idx)
         )
      {
         if (block_ptr->orientation != 0)
            y0 += 80;
         tile_ptr  = block_ptr->tile;
         for(t=0; t<block_ptr->tiles_number; t++)
         {
            length = tile_ptr->data_length;
            ptr    = tile_ptr->data;
            x      = tile_ptr->xpos;
            y      = tile_ptr->ypos;
            if (tile_ptr->format == 0x0001)
               draw_sub_tile_isometric(dst, x0+x, y0+y, ptr, length);
            else
               draw_sub_tile_normal(dst, x0+x, y0+y, ptr, length);
            tile_ptr++;
         }
fprintf(stderr, "%c", b < 223 ? 32 + b : '*');
         return;
      }
   }
}
fprintf(stderr, ".");
}


// ==========================================================================
void new_gfx_mode(int x, int y)
{
   if (set_gfx_mode(GFX_AUTODETECT, x, y, 0, 0) < 0)
   {
      set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);
      printf("can't initialise the gfx mode %i %i 8 bpp\n", x, y);
      printf("%s\n", allegro_error);
      exit(1);
   }
}


// ==========================================================================
int main(int argc, char * argv[])
{
/*
   BLOCK_S    * block_ptr=NULL;
   SUB_TILE_S * tile_ptr=NULL;
   long       t, b;
   char       tmp[80];
   int        k, i, done = FALSE, scr=0, plotx=150, ploty=400,
              screen_x = 640, screen_y = 480, gridvalue=2, layer=0, box=1;
   char       d2palname[5][80] = {
                 {"d2pal\\act1.dat"},
                 {"d2pal\\act2.dat"},
                 {"d2pal\\act3.dat"},
                 {"d2pal\\act4.dat"},
                 {"d2pal\\act5.dat"}
              };
   char       * pal_name = NULL;;
*/
   BITMAP * savebmp = NULL, * tmpbmp = NULL;
   FILE   * in;
   int    t, screen_x = 1024, screen_y = 768, c, maxlayer, pcx_w, pcx_h,
          xx, yy, base_x, base_y, prm_pal = FALSE;
//   long   w, h, x1, act, x2, x3, x4, f, i, i2, n, n2, p, old_pos;
   long   x1, w, h, act, x2, f, i, x3, x4, x, y, n, wa_st, fl_st;
   char   p1, p2, p3, p4, dt1_names[32][256];

   
   // allegro stuff
   allegro_init();
   set_color_depth(8);
   atexit(ds1layer_exit);

   // init
   for (i=0; i<32; i++)
   {
      glb_block[i] = NULL;
      glb_block_number[i] = 0;
   }
   
   // header & syntaxe
   printf("DS1LAYER v 0.25, by Paul Siramy, Freeware\n"
          "=========================================\n");
   if (argc < 3)
   {
      printf("syntaxe : dt1layer <file.ds1> [-pal <file.dat>] <file1.dt1> ... <fileN.dt1>\n");
      exit(0);
   }

   // palette & atexit()
   if (strcmp(argv[2], "-pal") == 0)
   {
      load_palette(argv[3]);
      prm_pal = TRUE;
   }
   else
   {
      printf("opening default palette\n   ");
      load_palette("d2pal\\act1.dat");
   }

   // dt1 file
   glb_dt1_file = 0;
   for (i = prm_pal==TRUE ? 4 : 2; i<argc; i++)
   {
      in = fopen(argv[i], "rb");
      if (in == NULL)
      {
         printf("can't open %s\n", argv[i]);
         exit(1);
      }
      else
         printf("opening %s\n", argv[i]);
      read_dt1_header(in, glb_dt1_file);
      read_blocks_headers(in, glb_dt1_file);
      for (t=0; t<glb_block_number[glb_dt1_file]; t++)
         read_tiles_of_block(in, t, glb_dt1_file);
      fclose(in);
      glb_dt1_file++;
   }

   // ds1 file
   in = fopen(argv[1], "rb");
   if (in == NULL)
   {
      printf("can't open %s\n", argv[1]);
      exit(1);
   }
   fread(&x1, 1, 4, in);
   printf("x1     = %li\n", x1);
   fread(&w, 1, 4, in);
   printf("width  = %li (+1 = %li)\n", w, w+1);
   fread(&h, 1, 4, in);
   printf("height = %li (+1 = %li)\n", h, h+1);
   if (x1 != 3)
   {
      fread(&act, 1, 4, in);
      printf("act    = %li ( = acte %li)\n", act, act+1);
   }
   else
      printf("no act data (x1 = 3)\n");
   if ((x1 != 3) && (x1 != 8))
   {
      fread(&x2, 1, 4, in);
      printf("x2     = %li\n", x2);
   }
   else
      printf("no x2 data (x1 = %i)\n", x1);
   fread(&f, 1, 4, in);
   printf("files  = %li\n", f);
   for (i=0; i<f; i++)
   {
      c = fgetc(in);
      while (c != 0)
         c = fgetc(in);
   }
   if ((x1 != 3) && (x1 != 8))
   {
      fread(&x3, 1, 4, in);
      printf("x3     = %li\n", x3);
      fread(&x4, 1, 4, in);
      printf("x4     = %li\n", x4);
   }
   else
   {
      printf("no x3 & no x4 data (x1 = %i)\n", x1);
   }

/*
   fl_st = x2 + (2 * x3);
   wa_st = x2;
*/
   fl_st = 2 * x3;
   wa_st = 0;

//   maxlayer = atoi(argv[2]);
   maxlayer = x2 + (x3 * 2) + x4;
printf("layer_to_read = %i\n", maxlayer);
   memset(layer, 0, 4 * 12 * 128 * 128);
   for (i=0; i < maxlayer; i++)
   {
      for (y=0; y<h+1; y++)
      {
         for (x=0; x<w+1; x++)
         {
            layer[i][y][x].p1 = fgetc(in);;
            layer[i][y][x].p2 = fgetc(in);;
            layer[i][y][x].p3 = fgetc(in);;
            layer[i][y][x].p4 = fgetc(in);;
         }
      }
   }
   fclose(in);

   // ds1 layer draw
   pcx_w = 80 * ((w+1) + (h+1));
   pcx_h = 40 * ((w+1) + (h+1));
fprintf(stderr, "pcx = %i * %i\n", pcx_w, pcx_h);
   tmpbmp = create_bitmap(pcx_w, pcx_h);
   clear(tmpbmp);
   for (y=0; y<h+1; y++)
   {
fprintf(stderr, "%i : ", y);
      base_x = (80 * h) - (80 * y);
      base_y = 40 * y;
      for (x=0; x<w+1; x++)
      {
         xx = base_x + (80 * x);
         yy = base_y + (40 * x);

         // floor 1
         if (layer[fl_st][y][x].p1 > 0)
         {
            draw_block_search1(tmpbmp, xx, yy,
               layer[fl_st][y][x].p2,     // sub-index
               (layer[fl_st][y][x].p3 / 16) + // main-index
               ((layer[fl_st][y][x].p4 & 0x0F) * 16)   // main index
            );
         }

         // floor 2
         if ((layer[fl_st+1][y][x].p1 > 0) && (x4 > 1) )
         {
            draw_block_search1(tmpbmp, xx, yy,
               layer[fl_st+1][y][x].p2,     // sub-index
               (layer[fl_st+1][y][x].p3 / 16) + // main-index
               ((layer[fl_st+1][y][x].p4 & 0x0F) * 16)   // main-index
            );
         }

         // wall 1
         if (layer[wa_st*2+1][y][x].p1 > 0)
         {
            if (layer[wa_st*2+1][y][x].p1 == 3)
            {
               draw_block_search2(tmpbmp, xx, yy,
                  3,                     // orientation
                  layer[wa_st*2][y][x].p2,     // sub-index
                  (layer[wa_st*2][y][x].p3 / 16) + // main-index
                  ((layer[wa_st*2][y][x].p4 & 0x0F) * 16)   // main-index
               );
               
               draw_block_search2(tmpbmp, xx, yy,
                  4,                     // orientation
                  layer[wa_st*2][y][x].p2,     // sub-index
                  (layer[wa_st*2][y][x].p3 / 16) + // main-index
                  ((layer[wa_st*2][y][x].p4 & 0x0F) * 16)   // main-index
               );
            }
            else // not upper / left corner
            {
               draw_block_search2(tmpbmp, xx, yy,
                  layer[wa_st*2+1][y][x].p1,     // orientation
                  layer[wa_st*2][y][x].p2,     // sub-index
                  (layer[wa_st*2][y][x].p3 / 16) +  // main-index
                  ((layer[wa_st*2][y][x].p4 & 0x0F) * 16)    // main-index
               );
            }
         }

         // wall 2
         if ( (layer[(wa_st+1)*2+1][y][x].p1 > 0) && (x3 > 1) )
         {
            if (layer[(wa_st+1)*2+1][y][x].p1 == 3)
            {
               draw_block_search2(tmpbmp, xx, yy,
                  3,                     // orientation
                  layer[(wa_st+1)*2][y][x].p2,     // sub-index
                  (layer[(wa_st+1)*2][y][x].p3 / 16) + // main-index
                  ((layer[(wa_st+1)*2][y][x].p4 & 0x0F) * 16) // main-index
               );
               
               draw_block_search2(tmpbmp, xx, yy,
                  4,                     // orientation
                  layer[(wa_st+1)*2][y][x].p2,     // sub-index
                  (layer[(wa_st+1)*2][y][x].p3 / 16) + // main-index
                  ((layer[(wa_st+1)*2][y][x].p4 & 0x0F) * 16)   // main-index
               );
            }
            else // not upper / left corner
            {
               draw_block_search2(tmpbmp, xx, yy,
                  layer[(wa_st+1)*2+1][y][x].p1,     // orientation
                  layer[(wa_st+1)*2][y][x].p2,     // sub-index
                  (layer[(wa_st+1)*2][y][x].p3 / 16) + // main-index
                  ((layer[(wa_st+1)*2][y][x].p3 & 0x0F) * 16)   // main-index
               );
            }
         }

      }
fprintf(stderr, "\n");
   }
   
   // end
   printf("saving pcx...");
fflush(stderr);
   save_pcx("ds1layer.pcx", tmpbmp, &the_pal);
   printf("done\n");
fflush(stderr);
   destroy_bitmap(tmpbmp);
   return 0;
}

